'''Unittests for feature methods.'''

import logging
from pathlib import Path

import unittest
import numpy as np
import pandas as pd
from pandas.api.types import is_string_dtype
from pandas.api.types import is_numeric_dtype

import ensembleset.feature_methods as fm
import tests.dummy_dataframe as test_data

Path('tests/logs').mkdir(parents=True, exist_ok=True)

logger = logging.getLogger()

logging.basicConfig(
    filename='tests/logs/test_feature_methods.log',
    filemode='w',
    level=logging.INFO,
    format='%(levelname)s - %(name)s - %(message)s'
)

class TestFeatureMethods(unittest.TestCase):
    '''Tests feature engineering method functions.'''

    def setUp(self):
        '''Dummy DataFrames for tests.'''

        self.dummy_df = test_data.DUMMY_DF


    def test_onehot_encoding(self):
        '''Tests string features onehot encoder.'''

        for testing_data in [None, self.dummy_df.copy()]:

            train_df, test_df=fm.onehot_encoding(
                train_df=self.dummy_df.copy(),
                test_df=testing_data,
                features=['strings'],
                kwargs={'sparse_output': False}
            )

            self.assertTrue(isinstance(train_df, pd.DataFrame))

            for feature in list(train_df.columns):
                self.assertTrue(is_numeric_dtype(train_df[feature]))

            if testing_data is not None:

                self.assertTrue(isinstance(test_df, pd.DataFrame))

                for feature in list(test_df.columns):
                    self.assertTrue(is_numeric_dtype(test_df[feature]))

            elif testing_data is None:
                self.assertEqual(test_df, None)


    def test_ordinal_encoding(self):
        '''Tests string feature ordinal encoder.'''

        for testing_data in [None, self.dummy_df.copy()]:

            train_df, test_df=fm.ordinal_encoding(
                train_df=self.dummy_df.copy(),
                test_df=testing_data,
                features=['strings'],
                kwargs={
                    'handle_unknown': 'use_encoded_value',
                    'unknown_value': np.nan  
                }
            )

            self.assertTrue(isinstance(train_df, pd.DataFrame))
            self.assertFalse(is_string_dtype(train_df['strings']))

            if testing_data is not None:
                self.assertTrue(isinstance(test_df, pd.DataFrame))
                self.assertFalse(is_string_dtype(test_df['strings']))


    def test_poly_features(self):
        '''Tests polynomial feature transformer.'''

        for testing_data in [None, self.dummy_df.copy()]:

            train_df, test_df=fm.poly_features(
                train_df=self.dummy_df.copy(),
                test_df=testing_data,
                features=list(self.dummy_df.columns),
                kwargs={'degree': 2}
            )

            self.assertTrue(isinstance(train_df, pd.DataFrame))

            if testing_data is not None:
                self.assertTrue(isinstance(test_df, pd.DataFrame))


    def test_spline_features(self):
        '''Tests spline features transformer.'''

        for testing_data in [None, self.dummy_df.copy()]:

            train_df, test_df=fm.spline_features(
                train_df=self.dummy_df.copy(),
                test_df=testing_data,
                features=list(self.dummy_df.columns),
                kwargs={'n_knots': 2}
            )

            self.assertTrue(isinstance(train_df, pd.DataFrame))

            if testing_data is not None:
                self.assertTrue(isinstance(test_df, pd.DataFrame))


    def test_log_features(self):
        '''Tests log features transformer.'''

        for testing_data in [None, self.dummy_df.copy()]:

            train_df, test_df=fm.log_features(
                train_df=self.dummy_df.copy(),
                test_df=testing_data,
                features=list(self.dummy_df.columns),
                kwargs={'base': '2'}
            )

            self.assertTrue(isinstance(train_df, pd.DataFrame))

            if testing_data is not None:
                self.assertTrue(isinstance(test_df, pd.DataFrame))

        for testing_data in [None, self.dummy_df.copy()]:

            train_df, test_df=fm.log_features(
                train_df=self.dummy_df.copy(),
                test_df=testing_data,
                features=list(self.dummy_df.columns),
                kwargs={'base': 'e'}
            )

            self.assertTrue(isinstance(train_df, pd.DataFrame))

            if testing_data is not None:
                self.assertTrue(isinstance(test_df, pd.DataFrame))

        for testing_data in [None, self.dummy_df.copy()]:

            train_df, test_df=fm.log_features(
                train_df=self.dummy_df.copy(),
                test_df=testing_data,
                features=list(self.dummy_df.columns),
                kwargs={'base': '10'}
            )

            self.assertTrue(isinstance(train_df, pd.DataFrame))

            if testing_data is not None:
                self.assertTrue(isinstance(test_df, pd.DataFrame))


    def test_ratio_features(self):
        '''Tests ratio feature transformer.'''

        for testing_data in [None, self.dummy_df.copy()]:

            train_df, test_df=fm.ratio_features(
                self.dummy_df.copy(),
                testing_data,
                list(self.dummy_df.columns),
                {'div_zero_value': np.nan}
            )

        self.assertTrue(isinstance(train_df, pd.DataFrame))

        if testing_data is not None:
            self.assertTrue(isinstance(test_df, pd.DataFrame))


    def test_exponential_features(self):
        '''Tests exponential features transformer.'''

        for testing_data in [None, self.dummy_df.copy()]:

            train_df, test_df=fm.exponential_features(
                self.dummy_df.copy(),
                testing_data,
                list(self.dummy_df.columns),
                {'base': 'e'}
            )

        self.assertTrue(isinstance(train_df, pd.DataFrame))

        if testing_data is not None:
            self.assertTrue(isinstance(test_df, pd.DataFrame))


        for testing_data in [None, self.dummy_df.copy()]:

            train_df, test_df=fm.exponential_features(
                self.dummy_df.copy(),
                testing_data,
                list(self.dummy_df.columns),
                {'base': '2'}
            )

        self.assertTrue(isinstance(train_df, pd.DataFrame))

        if testing_data is not None:
            self.assertTrue(isinstance(test_df, pd.DataFrame))


    def test_sum_features(self):
        '''Tests sum features transformer.'''

        for testing_data in [None, self.dummy_df.copy()]:

            train_df, test_df=fm.sum_features(
                self.dummy_df.copy(),
                testing_data,
                list(self.dummy_df.columns),
                {'n_addends': 2}
            )

            self.assertTrue(isinstance(train_df, pd.DataFrame))

            if testing_data is not None:
                self.assertTrue(isinstance(test_df, pd.DataFrame))

        for testing_data in [None, self.dummy_df.copy()]:

            train_df, test_df=fm.sum_features(
                self.dummy_df.copy(),
                testing_data,
                list(self.dummy_df.columns),
                {'n_addends': 3}
            )

        self.assertTrue(isinstance(train_df, pd.DataFrame))

        if testing_data is not None:
            self.assertTrue(isinstance(test_df, pd.DataFrame))


    def test_difference_features(self):
        '''Tests difference features transformer.'''

        for testing_data in [None, self.dummy_df.copy()]:

            train_df, test_df=fm.difference_features(
                self.dummy_df.copy(),
                testing_data,
                list(self.dummy_df.columns),
                {'n_subtrahends': 2}
            )

            self.assertTrue(isinstance(train_df, pd.DataFrame))

            if testing_data is not None:
                self.assertTrue(isinstance(test_df, pd.DataFrame))

        for testing_data in [None, self.dummy_df.copy()]:

            train_df, test_df=fm.difference_features(
                self.dummy_df.copy(),
                testing_data,
                list(self.dummy_df.columns),
                {'n_subtrahends': 100}
            )

        self.assertTrue(isinstance(train_df, pd.DataFrame))

        if testing_data is not None:
            self.assertTrue(isinstance(test_df, pd.DataFrame))


    def test_kde_smoothing(self):
        '''Tests kde smoother.'''

        train_df, test_df=fm.kde_smoothing( # pylint: disable=E1101
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            list(self.dummy_df.columns),
            {'bandwidth': 'silverman', 'sample_size': 1000}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))


    def test_kbins_quantization(self):
        '''Tests kbins discretizer.'''

        train_df, test_df=fm.kbins_quantization( # pylint: disable=E1101
            self.dummy_df.copy(),
            self.dummy_df.copy(),
            list(self.dummy_df.columns),
            {'n_bins': 64, 'encode': 'ordinal', 'strategy': 'quantile'}
        )

        self.assertTrue(isinstance(train_df, pd.DataFrame))
        self.assertTrue(isinstance(test_df, pd.DataFrame))



'''Generates variations of a dataset using a pool of feature engineering
techniques. Used for training ensemble models.'''

import logging
from pathlib import Path
from random import choice, shuffle

import h5py
import numpy as np
import pandas as pd

import ensembleset.feature_engineerings as engineerings
import ensembleset.feature_methods as fm

class DataSet:
    '''Dataset generator class.'''

    def __init__(
            self,
            label: str,
            train_data: pd.DataFrame,
            test_data: pd.DataFrame = None,
            string_features: list = None,
            data_directory: str = 'ensembleset_data'
        ):

        logger = logging.getLogger(__name__)
        logger.addHandler(logging.NullHandler())

        # Check user argument types
        type_check = self._check_argument_types(
            label=label,
            train_data=train_data,
            test_data=test_data,
            string_features=string_features,
            data_directory=data_directory
        )

        # If the type check passed, assign arguments to attributes
        if type_check is True:
            self.label = label
            self.train_data = train_data.copy()

            if test_data is not None:
                self.test_data = test_data.copy()

            else:
                self.test_data = None

            self.string_features = string_features
            self.data_directory = data_directory

        # Get the init logger
        logger.info("Training label: '%s'", self.label)
        logger.info('Training data: %s', type(self.train_data))
        logger.info('Testing data: %s', type(self.test_data))
        logger.info('String features: %s', self.string_features)
        logger.info('Data directory: %s', self.data_directory)

        # Enforce string type on DataFrame columns
        self.train_data.columns = self.train_data.columns.astype(str)

        if self.test_data is not None:
            self.test_data.columns = self.test_data.columns.astype(str)

        # Retrieve and assign the training labels, set NAN if they don't exist
        # then remove them from the training data
        if self.label in self.train_data.columns:
            self.train_labels=np.array(self.train_data[label])
            self.train_data.drop(self.label, axis=1, inplace=True)

        else:
            self.train_labels=[np.nan] * len(self.train_data)

        # Retrieve and assign the testing labels, set NAN if they don't exist
        # then remove them from the training data
        if self.test_data is not None:
            if self.label in self.test_data.columns:
                self.test_labels=np.array(self.test_data[label])
                self.test_data.drop(self.label, axis=1, inplace=True)

            else:
                self.test_labels=[np.nan] * len(self.test_data)

        else:
            self.test_labels = None

        # Create the HDF5 output
        Path(self.data_directory).mkdir(parents=True, exist_ok=True)

        # Create groups for training and testing datasets
        with h5py.File(f'{self.data_directory}/dataset.h5', 'a') as hdf:

            _ = hdf.require_group('train')

            if self.test_data is not None:
                _ = hdf.require_group('test')

        # Add the training and testing labels
        with h5py.File(f'{self.data_directory}/dataset.h5', 'w') as hdf:

            _ = hdf.create_dataset('train/labels', data=self.train_labels)

            if self.test_data is not None:
                _ = hdf.create_dataset('test/labels', data=self.test_labels)

        # Define the feature engineering pipeline methods
        self.string_encodings=engineerings.STRING_ENCODINGS
        self.numerical_methods=engineerings.NUMERICAL_METHODS


    def make_datasets(self, n_datasets:int, frac_features:int, n_steps:int):
        '''Makes n datasets with different feature subsets and pipelines.'''

        logger = logging.getLogger(__name__ + '.make_datasets')
        logger.addHandler(logging.NullHandler())

        logger.info('Will make %s datasets', n_datasets)
        logger.info('Running %s feature engineering steps per dataset', n_steps)
        logger.info('Selecting %s percent of features for each step', round(frac_features * 100))

        with h5py.File(f'{self.data_directory}/dataset.h5', 'a') as hdf:

            # Generate n datasets
            for n in range(n_datasets):

                logger.info('Generating dataset %s of %s', n+1, n_datasets)
                logger.info('Input training data shape: %s', self.train_data.shape)

                # Take a copy of the training and test data
                train_df = self.train_data.copy()

                if self.test_data is not None:
                    test_df = self.test_data.copy()
                    logger.info('Input testing data shape: %s', self.test_data.shape)

                else:
                    test_df = None

                # Generate a data pipeline
                pipeline = self._generate_data_pipeline(n_steps)

                # Loop on and apply each method in the pipeline
                for method, arguments in pipeline.items():

                    func = getattr(fm, method)

                    if method in self.string_encodings:

                        logger.info('Applying %s to %s' , method, self.string_features)

                        train_df, test_df = func(
                            train_df,
                            test_df,
                            self.string_features,
                            arguments
                        )

                    else:

                        n_features = int(len(train_df.columns.to_list()) * frac_features)
                        n_features = max([n_features, 1])
                        features = self._select_features(n_features, train_df)

                        logger.info('Applying %s to %s features' , method, len(features))

                        train_df, test_df = func(
                            train_df,
                            test_df,
                            features,
                            arguments
                        )

                        logger.info('New training data shape: %s', train_df.shape)

                        if test_df is not None:
                            logger.info('New testing data shape: %s', test_df.shape)

                # Save the results to HDF5 output
                _ = hdf.create_dataset(f'train/{n}', data=np.array(train_df).astype(np.float64))

                if test_df is not None:
                    _ = hdf.create_dataset(f'test/{n}', data=np.array(test_df).astype(np.float64))


    def _select_features(self, n_features:int, data_df:pd.DataFrame):
        '''Selects a random subset of features.'''

        features = data_df.columns.to_list()
        shuffle(features)
        features = features[:n_features]

        return features


    def _generate_data_pipeline(self, n_steps:int):
        '''Generates one random sequence of feature engineering operations. Starts with
        a string encoding method if we have string features.'''

        pipeline={}

        # Choose a string encoding method, if needed
        if self.string_features is not None:
            options = list(self.string_encodings.keys())
            selection = choice(options)
            pipeline[selection] = self.string_encodings[selection]

        # Construct a random sequence of numerical feature engineering methods
        methods = list(self.numerical_methods.keys())
        shuffle(methods)
        methods = methods[:n_steps]

        for method in methods:

            pipeline[method] = {}
            parameters = self.numerical_methods[method]

            for parameter, values in parameters.items():

                value = choice(values)
                pipeline[method][parameter] = value

        return pipeline


    def _check_argument_types(
            self,
            label: str,
            train_data: pd.DataFrame,
            test_data: pd.DataFrame,
            string_features: list,
            data_directory: str
    ) -> bool:

        '''Checks user argument types, returns true or false for all passing.'''

        check_pass = False

        if isinstance(label, str):
            check_pass = True

        else:
            raise TypeError('Label is not a string.')

        if isinstance(train_data, pd.DataFrame):
            check_pass = True

        else:
            raise TypeError('Train data is not a Pandas DataFrame.')

        if isinstance(test_data, pd.DataFrame) or test_data is None:
            check_pass = True

        else:
            raise TypeError('Test data is not a Pandas DataFrame.')

        if isinstance(string_features, list) or string_features is None:
            check_pass = True

        else:
            raise TypeError('String features is not a list.')

        if isinstance(data_directory, str):
            check_pass = True

        else:
            raise TypeError('Data directory is not a string.')

        return check_pass



'''Plotting functions for notebooks.'''

import matplotlib.pyplot as plt


def model_eval(plot_title, feature_name, predictions, labels):
    '''Plots true vs predicted values and residuals vs
    true values'''

    fig, axs = plt.subplots(1, 2, figsize=(8, 4))

    fig.suptitle(plot_title)

    axs[0].set_title('Test set predictions')
    axs[0].scatter(labels, predictions, color='black', s=0.5)
    axs[0].set_xlabel(f'true {feature_name}')
    axs[0].set_ylabel(f'predicted {feature_name}')

    axs[1].set_title('Test set residuals')
    axs[1].scatter(labels, labels - predictions, color='black', s=0.5)
    axs[1].set_xlabel(f'true {feature_name}')
    axs[1].set_ylabel(f'{feature_name} (true - predicted)')

    fig.tight_layout()



'''Collection of functions to run feature engineering operations.'''

import logging
import multiprocessing as mp
from math import e
from itertools import permutations, combinations
from typing import Tuple

import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype
from scipy.stats import gaussian_kde
from sklearn.exceptions import ConvergenceWarning
from sklearn.impute import KNNImputer
from sklearn.preprocessing import (
    MinMaxScaler,
    OneHotEncoder,
    OrdinalEncoder,
    PolynomialFeatures,
    SplineTransformer,
    KBinsDiscretizer
)

pd.set_option('display.width', 100)
pd.set_option('display.max_rows', 100)
pd.set_option('display.max_columns', 100)


def onehot_encoding(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Runs sklearn's one hot encoder.'''

    logger = logging.getLogger(__name__ + '.onehot_encoding')
    logger.addHandler(logging.NullHandler())
    logger.info('One-hot encoding string features')

    if features is not None:

        encoder=OneHotEncoder(**kwargs)

        encoded_data=encoder.fit_transform(train_df[features])
        encoded_df=pd.DataFrame(encoded_data, columns=encoder.get_feature_names_out())
        train_df=pd.concat(
            [train_df.reset_index(drop=True), encoded_df.reset_index(drop=True)],
            axis=1
        )
        train_df.drop(features, axis=1, inplace=True)

        if test_df is not None:
            encoded_data=encoder.transform(test_df[features])
            encoded_df=pd.DataFrame(encoded_data, columns=encoder.get_feature_names_out())
            test_df=pd.concat(
                [test_df.reset_index(drop=True), encoded_df.reset_index(drop=True)],
                axis=1
            )
            test_df.drop(features, axis=1, inplace=True)

    return train_df, test_df


def ordinal_encoding(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Runs sklearn's ordinal encoder.'''

    logger = logging.getLogger(__name__ + '.ordinal_encoding')
    logger.addHandler(logging.NullHandler())
    logger.info('Ordinal encoding string features')

    if features is not None:

        encoder=OrdinalEncoder(**kwargs)

        train_df[features]=encoder.fit_transform(train_df[features])

        if test_df is not None:
            test_df[features]=encoder.transform(test_df[features])

    return train_df, test_df


def poly_features(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Runs sklearn's polynomial feature transformer.'''

    logger = logging.getLogger(__name__ + '.poly_features')
    logger.addHandler(logging.NullHandler())
    logger.info('Adding polynomial features')

    features, train_working_df, test_working_df=preprocess_features(
        features=features,
        train_df=train_df,
        test_df=test_df,
        preprocessing_steps=[
            'exclude_string_features',
            'enforce_floats',
            'remove_inf', 
            'remove_large_nums',
            'remove_small_nums',
            'knn_impute',
            'remove_constants',
            'scale_to_range'
        ]
    )

    if features is not None:

        transformer=PolynomialFeatures(**kwargs)

        try:
            transformed_data=transformer.fit_transform(train_working_df[features])
            new_columns=transformer.get_feature_names_out()
            transformed_train_df=pd.DataFrame(transformed_data, columns=new_columns)

        except ValueError:
            print('Value error in poly feature transformer.')

        transformed_test_df = None

        if test_df is not None:

            try:
                transformed_data=transformer.transform(test_working_df[features])
                new_columns=transformer.get_feature_names_out()
                transformed_test_df=pd.DataFrame(transformed_data, columns=new_columns)

            except ValueError:
                print('Value error in poly feature transformer.')

        train_df, test_df = add_new_features(
            new_train_features = transformed_train_df,
            new_test_features = transformed_test_df,
            train_df = train_df,
            test_df = test_df
        )

    return train_df, test_df


def spline_features(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Runs sklearn's polynomial feature transformer.'''

    logger = logging.getLogger(__name__ + '.spline_features')
    logger.addHandler(logging.NullHandler())
    logger.info('Adding spline features')

    features, train_working_df, test_working_df=preprocess_features(
        features=features,
        train_df=train_df,
        test_df=test_df,
        preprocessing_steps=[
            'exclude_string_features',
            'enforce_floats',
            'remove_inf',
            'remove_large_nums',
            'remove_small_nums',
            'knn_impute',
            'remove_constants',
            'scale_to_range'
        ]
    )

    if features is not None:

        transformer=SplineTransformer(**kwargs)

        try:
            transformed_data=transformer.fit_transform(train_working_df[features])
            new_columns=transformer.get_feature_names_out()
            transformed_train_df=pd.DataFrame(transformed_data, columns=new_columns)

        except ValueError:
            print('Caught value error error during spline feature concatenation')

        transformed_test_df = None

        if test_df is not None:

            try:
                transformed_data=transformer.transform(test_working_df[features])
                new_columns=transformer.get_feature_names_out()
                transformed_test_df=pd.DataFrame(transformed_data, columns=new_columns)

            except ValueError:
                print('Caught value error error during spline feature concatenation')

        train_df, test_df = add_new_features(
            new_train_features = transformed_train_df,
            new_test_features = transformed_test_df,
            train_df = train_df,
            test_df = test_df
        )

    return train_df, test_df


def log_features(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Takes log of feature, uses sklearn min-max scaler if needed
    to avoid undefined log errors.'''

    logger = logging.getLogger(__name__ + '.log_features')
    logger.addHandler(logging.NullHandler())
    logger.info('Adding log features')

    features, train_working_df, test_working_df = preprocess_features(
        features=features,
        train_df=train_df,
        test_df=test_df,
        preprocessing_steps=[
            'exclude_string_features',
            'enforce_floats',
            'remove_inf', 
            'remove_large_nums',
            'remove_small_nums',
            'knn_impute',
            'remove_constants'
        ]
    )

    features, train_working_df, test_working_df = scale_to_range(
        features=features,
        train_df=train_working_df,
        test_df=test_working_df,
        min_val=1,
        max_val=10
    )

    if features is not None:

        for feature in features:

            if kwargs['base'] == '2':
                train_df[f'{feature}_log2']=np.log2(train_working_df[feature])

                if test_df is not None:
                    test_df[f'{feature}_log2']=np.log2(test_working_df[feature])

            if kwargs['base'] == 'e':
                train_df[f'{feature}_ln']=np.log(train_working_df[feature])

                if test_df is not None:
                    test_df[f'{feature}_ln']=np.log(test_working_df[feature])

            if kwargs['base'] == '10':
                train_df[f'{feature}_log10']=np.log10(train_working_df[feature])

                if test_df is not None:
                    test_df[f'{feature}_log10']=np.log10(test_working_df[feature])

        train_df.dropna(axis=1, how='all', inplace=True)

        if test_df is not None:
            test_df.dropna(axis=1, how='all', inplace=True)

    return train_df, test_df


def ratio_features(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Adds every possible ratio feature, replaces divide by zero errors
    with np.nan.'''

    logger = logging.getLogger(__name__ + '.ratio_features')
    logger.addHandler(logging.NullHandler())
    logger.info('Adding ratio features')

    features, train_working_df, test_working_df=preprocess_features(
        features=features,
        train_df=train_df,
        test_df=test_df,
        preprocessing_steps=[
            'exclude_string_features',
            'enforce_floats',
            'remove_inf', 
            'remove_large_nums',
            'remove_small_nums',
            'knn_impute',
            'remove_constants'
        ]
    )

    if features is not None:

        features, train_working_df, test_working_df = scale_to_range(
            features=features,
            train_df=train_working_df,
            test_df=test_working_df,
            min_val=1,
            max_val=10
        )

        feature_pairs=permutations(features, 2)

        new_train_features={}
        new_test_features={}

        for feature_a, feature_b in feature_pairs:

            quotient = np.divide(
                np.array(train_working_df[feature_a]),
                np.array(train_working_df[feature_b]),
                out=np.array([kwargs['div_zero_value']]*len(train_working_df[feature_a])),
                where=np.array(train_working_df[feature_b]) != 0
            )

            new_train_features[f'{feature_a}_over_{feature_b}'] = quotient

            if test_df is not None:

                quotient = np.divide(
                    np.array(test_working_df[feature_a]),
                    np.array(test_working_df[feature_b]),
                    out=np.array([kwargs['div_zero_value']]*len(test_working_df[feature_a])),
                    where=np.array(test_working_df[feature_b]) != 0
                )

                new_test_features[f'{feature_a}_over_{feature_b}'] = quotient

        train_df, test_df=add_new_features(
            new_train_features = new_train_features,
            new_test_features = new_test_features,
            train_df = train_df,
            test_df = test_df
        )

    return train_df, test_df


def exponential_features(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Adds exponential features with base 2 or base e.'''

    logger = logging.getLogger(__name__ + '.exponential_features')
    logger.addHandler(logging.NullHandler())
    logger.info('Adding exponential features')

    features, train_working_df, test_working_df=preprocess_features(
        features=features,
        train_df=train_df,
        test_df=test_df,
        preprocessing_steps=[
            'exclude_string_features',
            'enforce_floats',
            'remove_inf', 
            'remove_large_nums',
            'remove_small_nums',
            'knn_impute',
            'remove_constants',
            'scale_to_range'
        ]
    )

    if features is not None:

        new_train_features={}
        new_test_features={}

        for feature in features:

            if kwargs['base'] == 'e':
                new_train_features[f'{feature}_exp_base_e'] = (
                    e**train_working_df[feature].astype(float)
                )

                if test_df is not None:
                    new_test_features[f'{feature}_exp_base_e'] = (
                        e**test_working_df[feature].astype(float)
                    )

            elif kwargs['base'] == '2':
                new_train_features[f'{feature}_exp_base_2'] = (
                    2**train_working_df[feature].astype(float))

                if test_df is not None:
                    new_test_features[f'{feature}_exp_base_2'] = (
                        2**test_working_df[feature].astype(float)
                    )

        train_df, test_df=add_new_features(
            new_train_features = new_train_features,
            new_test_features = new_test_features,
            train_df = train_df,
            test_df = test_df
        )

    return train_df, test_df


def sum_features(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Adds sum features for variable number of addends.'''

    logger = logging.getLogger(__name__ + '.sum_features')
    logger.addHandler(logging.NullHandler())
    logger.info('Adding sum features')

    features, train_working_df, test_working_df=preprocess_features(
        features=features,
        train_df=train_df,
        test_df=test_df,
        preprocessing_steps=[
            'exclude_string_features',
            'enforce_floats',
            'remove_inf', 
            'remove_large_nums',
            'remove_small_nums',
            'knn_impute',
            'remove_constants',
            'scale_to_range'
        ]
    )

    if features is not None:

        if kwargs['n_addends'] > len(features):
            n_addends=len(features)

        else:
            n_addends=kwargs['n_addends']

        new_test_features={}
        new_train_features={}
        addend_sets=list(combinations(features, n_addends))

        logger.info(
            'Will compute sums for %s sets of %s features', len(addend_sets), n_addends
        )

        for i, addend_set in enumerate(addend_sets):

            train_sum = [0]*len(train_working_df)

            for addend in addend_set:

                train_sum += train_working_df[addend]

            new_train_features[f'sum_feature_{i}'] = train_sum

            if test_df is not None:

                test_sum = [0]*len(test_working_df)

                for addend in addend_set:

                    test_sum += test_working_df[addend]

                new_test_features[f'sum_feature_{i}'] = test_sum

        train_df, test_df=add_new_features(
            new_train_features = new_train_features,
            new_test_features = new_test_features,
            train_df = train_df,
            test_df = test_df
        )

    else:
        logger.info('No features to sum')

    return train_df, test_df


def difference_features(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Adds difference features for variable number of subtrahends.'''

    logger = logging.getLogger(__name__ + '.difference_features')
    logger.addHandler(logging.NullHandler())
    logger.info('Adding difference features')

    features, train_working_df, test_working_df=preprocess_features(
        features=features,
        train_df=train_df,
        test_df=test_df,
        preprocessing_steps=[
            'exclude_string_features',
            'enforce_floats',
            'remove_inf', 
            'remove_large_nums',
            'remove_small_nums',
            'knn_impute',
            'remove_constants',
            'scale_to_range'
        ]
    )

    if features is not None:

        if kwargs['n_subtrahends'] > len(features):
            n_subtrahends=len(features)

        else:
            n_subtrahends=kwargs['n_subtrahends']

        new_test_features={}
        new_train_features={}
        subtrahend_sets=combinations(features, n_subtrahends)

        for subtrahend_set in subtrahend_sets:

            train_difference = train_working_df[subtrahend_set[0]]

            for subtrahend in subtrahend_set[1:]:

                train_difference -= train_working_df[subtrahend]

            new_train_features['-'.join(subtrahend_set)] = train_difference

            if test_df is not None:

                test_difference = test_working_df[subtrahend_set[0]]

                for subtrahend in subtrahend_set[1:]:

                    test_difference -= test_working_df[subtrahend]

                new_test_features['-'.join(subtrahend_set)] = test_difference

        train_df, test_df=add_new_features(
            new_train_features = new_train_features,
            new_test_features = new_test_features,
            train_df = train_df,
            test_df = test_df
        )

    return train_df, test_df


def kde_smoothing(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Uses kernel density estimation to smooth features.'''

    logger = logging.getLogger(__name__ + '.kde_smoothing')
    logger.addHandler(logging.NullHandler())
    logger.info('Adding kernel density estimate smoothed features')

    features, train_working_df, test_working_df=preprocess_features(
        features=features,
        train_df=train_df,
        test_df=test_df,
        preprocessing_steps=[
            'exclude_string_features',
            'enforce_floats',
            'remove_inf', 
            'remove_large_nums',
            'remove_small_nums',
            'knn_impute',
            'remove_constants',
            'scale_to_range'
        ]
    )

    if features is not None:

        new_test_features={}
        new_train_features={}

        if len(train_working_df) > kwargs['sample_size']:
            sample_df=train_working_df.sample(n=kwargs['sample_size'])

        else:
            sample_df=train_working_df

        workers = mp.cpu_count() - 2

        for feature in features:

            try:

                scipy_kde = gaussian_kde(
                    sample_df[feature].to_numpy().flatten(),
                    bw_method = kwargs['bandwidth']
                )

                with mp.Pool(workers) as p:
                    new_train_features[f'{feature}_kde'] = np.concatenate(p.map(
                        scipy_kde,
                        np.array_split(train_working_df[feature].to_numpy().flatten(), workers)
                    ))

                if test_df is not None:
                    with mp.Pool(workers) as p:
                        new_test_features[f'{feature}_kde'] = np.concatenate(p.map(
                            scipy_kde,
                            np.array_split(test_working_df[feature].to_numpy().flatten(), workers)
                        ))

            except np.linalg.LinAlgError:
                print('Numpy linear algebra error in gaussian KDE.')

        train_df, test_df=add_new_features(
            new_train_features = new_train_features,
            new_test_features = new_test_features,
            train_df = train_df,
            test_df = test_df
        )

    return train_df, test_df


def kbins_quantization(
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        features:list,
        kwargs:dict
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Discretizes feature with Kbins quantization.'''

    logger = logging.getLogger(__name__ + '.kbins_quantization')
    logger.addHandler(logging.NullHandler())
    logger.info('Adding k-bins quantized features')

    features, train_working_df, test_working_df=preprocess_features(
        features=features,
        train_df=train_df,
        test_df=test_df,
        preprocessing_steps=[
            'exclude_string_features',
            'enforce_floats',
            'remove_inf', 
            'remove_large_nums',
            'remove_small_nums',
            'knn_impute',
            'remove_constants',
            'scale_to_range'
        ]
    )

    if features is not None:

        if len(train_df) <= kwargs['n_bins']:
            kwargs['n_bins'] = len(train_df) - 1

        kbins = KBinsDiscretizer(**kwargs)

        try:
            binned_features = kbins.fit_transform(train_working_df[features])
            binned_feature_names = kbins.get_feature_names_out()
            binned_feature_names = [f'{feature_name}_bins' for feature_name in binned_feature_names]
            binned_train_features_df = pd.DataFrame(binned_features, columns=binned_feature_names)

        except ConvergenceWarning:
            print('Caught ConvergenceWarning in KbinsDescretizer.')

        except UserWarning:
            print('Caught UserWarning in KbinsDiscretizer.')

        except ValueError:
            print('Caught ValueError in KbinsDiscretizer.')

        binned_test_features_df = None

        if test_df is not None:

            try:
                binned_features = kbins.transform(test_working_df[features])
                binned_feature_names = kbins.get_feature_names_out()
                binned_feature_names = (
                    [f'{feature_name}_bins' for feature_name in binned_feature_names]
                )
                binned_test_features_df = pd.DataFrame(
                    binned_features,
                    columns=binned_feature_names
                )

            except ConvergenceWarning:
                print('Caught ConvergenceWarning in KbinsDescretizer.')

            except UserWarning:
                print('Caught UserWarning in KbinsDiscretizer.')

            except ValueError:
                print('Caught ValueError in KbinsDiscretizer.')

        train_df, test_df = add_new_features(
            new_train_features = binned_train_features_df,
            new_test_features = binned_test_features_df,
            train_df = train_df,
            test_df = test_df
        )

    return train_df, test_df


def preprocess_features(
        features: list,
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        preprocessing_steps:list
) -> Tuple[list, pd.DataFrame, pd.DataFrame]:

    '''Runs feature preprocessing steps.'''

    logger = logging.getLogger(__name__ + '.preprocess_features')
    logger.addHandler(logging.NullHandler())

    train_working_df=train_df.copy()

    if test_df is not None:
        test_working_df = test_df.copy()

    else:
        test_working_df = None

    for preprocessing_step in preprocessing_steps:

        preprocessing_func = globals().get(preprocessing_step)

        if features is not None:

            logger.info('Preprocessor running %s', preprocessing_step)

            features, train_working_df, test_working_df = preprocessing_func(
                features,
                train_working_df,
                test_working_df
            )

        else:

            logger.info('Preprocessor step %s received no features', preprocessing_step)


    return features, train_working_df, test_working_df


def exclude_string_features(
        features: list,
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
) -> Tuple[list, pd.DataFrame, pd.DataFrame]:

    '''Removes string features from features list.'''

    if features is not None:

        for feature in features:
            if test_df is not None:
                if (is_numeric_dtype(train_df[feature]) is False or
                    is_numeric_dtype(test_df[feature]) is False):

                    train_df.drop(feature, axis=1, inplace=True, errors='ignore')
                    test_df.drop(feature, axis=1, inplace=True, errors='ignore')
                    features.remove(feature)

            elif test_df is None:
                if is_numeric_dtype(train_df[feature]) is False:
                    train_df.drop(feature, axis=1, inplace=True, errors='ignore')
                    features.remove(feature)

    return features, train_df, test_df


def enforce_floats(
        features: list,
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
) -> Tuple[list, pd.DataFrame, pd.DataFrame]:

    '''Changes features to float dtype.'''

    if features is not None:

        train_df[features]=train_df[features].astype(float).copy()

        if test_df is not None:
            test_df[features]=test_df[features].astype(float).copy()

    return features, train_df, test_df


def remove_inf(
        features: list,
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
) -> Tuple[list, pd.DataFrame, pd.DataFrame]:

    '''Replaces any np.inf values with np.NAN.'''

    if features is not None:

        # Get rid of np.inf
        train_df[features]=train_df[features].replace(
            [np.inf, -np.inf],
            np.nan
        )

        if test_df is not None:
            test_df[features]=test_df[features].replace(
                [np.inf, -np.inf],
                np.nan
            )

    return features, train_df, test_df


def remove_large_nums(
        features: list,
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
) -> Tuple[list, pd.DataFrame, pd.DataFrame]:

    '''Replaces numbers larger than the cube root of the float64 limit with np.nan.'''

    if features is not None:

        # Get rid of large values
        train_df[features] = train_df[features].mask(
            abs(train_df[features]) > 1.0*10**102
        )

        if test_df is not None:
            test_df[features] = test_df[features].mask(
                test_df[features] > 1.0*10**102
            )

    return features, train_df, test_df


def remove_small_nums(
        features: list,
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
) -> Tuple[list, pd.DataFrame, pd.DataFrame]:

    '''Replaces values smaller than the float64 limit with zero.'''

    if features is not None:

        # Get rid of small values
        train_df[features] = train_df[features].mask(
            abs(train_df[features]) < 1.0-102
        ).fillna(0.0)

        if test_df is not None:
            test_df[features] = test_df[features].mask(
                abs(test_df[features]) < 1.0-102
            ).fillna(0.0)

    return features, train_df, test_df


def knn_impute(
        features: list,
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
) -> Tuple[list, pd.DataFrame, pd.DataFrame]:

    '''Uses SciKit-lean's KNN imputer to fill np.nan.'''

    if features is not None:

        imputer=KNNImputer()
        train_df[features] = imputer.fit_transform(train_df[features])

        if test_df is not None:
            test_df[features] = imputer.transform(test_df[features])

    return features, train_df, test_df


def remove_constants(
        features: list,
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
) -> Tuple[list, pd.DataFrame, pd.DataFrame]:

    '''Removes constant valued features.'''

    constant_features=train_df.loc[:,train_df.nunique(dropna=False) == 1]
    train_df.drop(constant_features, axis=1, inplace=True)

    if test_df is not None:
        test_df.drop(constant_features, axis=1, inplace=True)

    new_features = list(set(features) & set(train_df.columns.to_list()))

    if len(new_features) == 0:
        new_features = None
        train_df = None
        test_df = None

    return new_features, train_df, test_df


def scale_to_range(
        features: list,
        train_df:pd.DataFrame,
        test_df:pd.DataFrame,
        min_val:float = 0.0,
        max_val:float = 1.0
) -> Tuple[list, pd.DataFrame, pd.DataFrame]:

    '''Scales features into range'''

    if features is not None:

        scaler=MinMaxScaler(feature_range=(min_val, max_val))
        train_df[features]=scaler.fit_transform(train_df[features])

        if test_df is not None:
            test_df[features]=scaler.transform(test_df[features])

    return features, train_df, test_df


def add_new_features(
    new_train_features,
    new_test_features,
    train_df:pd.DataFrame,
    test_df:pd.DataFrame
) -> Tuple[pd.DataFrame, pd.DataFrame]:

    '''Adds new features to dataframes'''

    if isinstance(new_test_features, dict):
        new_train_features=pd.DataFrame.from_dict(new_train_features)

        if new_test_features is not None:
            new_test_features=pd.DataFrame.from_dict(new_test_features)

    train_df=pd.concat(
        [train_df.reset_index(drop=True), new_train_features.reset_index(drop=True)],
        axis=1
    )

    train_df = train_df.loc[:,~train_df.columns.duplicated()].copy()
    train_df.sort_index(axis=1, inplace=True)
    train_df.reset_index(inplace=True, drop=True)

    if test_df is not None:
        test_df=pd.concat(
            [test_df.reset_index(drop=True), new_test_features.reset_index(drop=True)],
            axis=1
        )

        test_df = test_df.loc[:,~test_df.columns.duplicated()].copy()
        test_df.sort_index(axis=1, inplace=True)
        test_df.reset_index(inplace=True, drop=True)

    return train_df, test_df
